@@ -43,9 +43,9 @@
 #define AGENT_PATH "/org/bluez/agent/sendto"
 
 static DBusGConnection *conn = NULL;
+static ObexAgent *agent = NULL;
 
 static GtkWidget *dialog;
-static GtkWidget *button;
 static GtkWidget *label_from;
 static GtkWidget *image_status;
 static GtkWidget *label_status;
@@ -55,6 +55,7 @@ static gchar *option_device = NULL;
 static gchar *option_device_name = NULL;
 static gchar **option_files = NULL;
 
+static DBusGProxy *current_transfer = NULL;
 static guint64 current_size = 0;
 static guint64 total_size = 0;
 static guint64 total_sent = 0;
@@ -65,6 +66,29 @@ static int file_index = 0;
 static gint64 first_update = 0;
 static gint64 last_update = 0;
 
+static void value_free(GValue *value)
+{
+	g_value_unset(value);
+	g_free(value);
+}
+
+static GHashTable *
+new_hash_table_for_address (const char *address)
+{
+	GHashTable *hash;
+	GValue *value;
+
+	hash = g_hash_table_new_full(g_str_hash, g_str_equal,
+				     g_free, (GDestroyNotify) value_free);
+
+	value = g_new0(GValue, 1);
+	g_value_init(value, G_TYPE_STRING);
+	g_value_set_string(value, address);
+	g_hash_table_insert(hash, g_strdup("Destination"), value);
+
+	return hash;
+}
+
 static gchar *filename_to_path(const gchar *filename)
 {
 	GFile *file;
@@ -125,9 +149,27 @@ static gchar *format_time(gint seconds)
 				"approximately %'d hours", hours), hours);
 }
 
+static void
+set_response_visible (GtkDialog *dialog,
+		      int response_id,
+		      gboolean visible)
+{
+	GtkWidget *widget;
+
+	widget = gtk_dialog_get_widget_for_response (dialog, response_id);
+	gtk_widget_set_no_show_all (widget, TRUE);
+	gtk_widget_set_visible (widget, visible);
+}
+
 static void response_callback(GtkWidget *dialog,
 					gint response, gpointer user_data)
 {
+	if (current_transfer != NULL) {
+		obex_agent_set_error_func(agent, NULL, NULL);
+		dbus_g_proxy_call_no_reply (current_transfer, "Cancel", G_TYPE_INVALID);
+		g_object_unref (current_transfer);
+		current_transfer = NULL;
+	}
 	gtk_widget_destroy(dialog);
 
 	gtk_main_quit();
@@ -149,8 +191,10 @@ static void create_window(void)
 
 	dialog = gtk_dialog_new_with_buttons(_("File Transfer"), NULL,
 				GTK_DIALOG_NO_SEPARATOR,
+				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+				GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
 				NULL);
-	button = gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
+	set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, FALSE);
 	gtk_window_set_type_hint(GTK_WINDOW(dialog),
 						GDK_WINDOW_TYPE_HINT_NORMAL);
 	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
@@ -187,9 +231,6 @@ static void create_window(void)
 	label_from = gtk_label_new(NULL);
 	gtk_misc_set_alignment(GTK_MISC(label_from), 0, 0.5);
 	gtk_label_set_ellipsize(GTK_LABEL(label_from), PANGO_ELLIPSIZE_MIDDLE);
-	text = g_get_current_dir();
-	gtk_label_set_markup(GTK_LABEL(label_from), text);
-	g_free(text);
 	gtk_table_attach_defaults(GTK_TABLE(table), label_from, 1, 2, 0, 1);
 
 	label = gtk_label_new(NULL);
@@ -237,7 +278,6 @@ static void create_window(void)
 static gchar *get_error_message(GError *error)
 {
 	char *message;
-	const gchar *name;
 
 	if (error == NULL)
 		return g_strdup(_("An unknown error occurred"));
@@ -247,8 +287,7 @@ static gchar *get_error_message(GError *error)
 		goto done;
 	}
 
-	name = dbus_g_error_get_name(error);
-	if (g_str_equal(name, OPENOBEX_CONNECTION_FAILED) == TRUE &&
+	if (dbus_g_error_has_name(error, OPENOBEX_CONNECTION_FAILED) == TRUE &&
 					is_palm_device(option_device)) {
 		message = g_strdup(_("Make sure that remote device "
 					"is switched on and that it "
@@ -256,7 +295,10 @@ static gchar *get_error_message(GError *error)
 		goto done;
 	}
 
-	message = g_strdup(error->message);
+	if (*error->message == '\0')
+		message = g_strdup(_("An unknown error occurred"));
+	else
+		message = g_strdup(error->message);
 
 done:
 	g_error_free(error);
@@ -311,6 +353,7 @@ static void get_properties_callback (DBusGProxy *proxy,
 {
 	GError *error = NULL;
 	gchar *filename = option_files[file_index];
+	GFile *file, *dir;
 	gchar *basename, *text, *markup;
 	GHashTable *hash;
 
@@ -329,9 +372,19 @@ static void get_properties_callback (DBusGProxy *proxy,
 		g_hash_table_destroy(hash);
 	}
 
-	text = g_path_get_dirname(filename);
-	gtk_label_set_markup(GTK_LABEL(label_from), text);
-	g_free(text);
+	file = g_file_new_for_path (filename);
+	dir = g_file_get_parent (file);
+	g_object_unref (file);
+	if (g_file_has_uri_scheme (dir, "file") != FALSE) {
+		text = g_file_get_path (dir);
+	} else {
+		text = g_file_get_uri (dir);
+	}
+	markup = g_markup_escape_text (text, -1);
+	g_free (text);
+	g_object_unref (dir);
+	gtk_label_set_markup(GTK_LABEL(label_from), markup);
+	g_free(markup);
 
 	basename = g_path_get_basename(filename);
 	text = g_strdup_printf(_("Sending %s"), basename);
@@ -350,10 +403,14 @@ static void get_properties_callback (DBusGProxy *proxy,
 static gboolean request_callback(DBusGMethodInvocation *context,
 				DBusGProxy *transfer, gpointer user_data)
 {
+	g_assert (current_transfer == NULL);
+
 	dbus_g_proxy_begin_call(transfer, "GetProperties",
 				get_properties_callback, NULL, NULL,
 				G_TYPE_INVALID);
 
+	current_transfer = g_object_ref (transfer);
+
 	dbus_g_method_return(context, "");
 
 	return TRUE;
@@ -425,7 +482,12 @@ static gboolean complete_callback(DBusGMethodInvocation *context,
 
 	file_index++;
 
-	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 1.0);
+	/* And we're done with the transfer */
+	g_object_unref (current_transfer);
+	current_transfer = NULL;
+
+	if (file_index == file_count)
+		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 1.0);
 
 	dbus_g_method_return(context);
 
@@ -446,6 +508,18 @@ static gboolean release_callback(DBusGMethodInvocation *context,
 	return TRUE;
 }
 
+static gboolean error_callback(DBusGMethodInvocation *context,
+			       DBusGProxy *transfer,
+			       const char *message,
+			       gpointer user_data)
+{
+	g_message ("Error message: %s", message);
+
+	dbus_g_method_return(context);
+
+	return TRUE;
+}
+
 static void send_notify(DBusGProxy *proxy,
 				DBusGProxyCall *call, void *user_data)
 {
@@ -460,9 +534,8 @@ static void send_notify(DBusGProxy *proxy,
 		gtk_label_set_markup(GTK_LABEL(label_status), message);
 		g_free (message);
 
-		gtk_button_set_label (GTK_BUTTON (button), GTK_STOCK_CLOSE);
-		gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
-						GTK_RESPONSE_CANCEL, TRUE);
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL, FALSE);
+		set_response_visible (GTK_DIALOG (dialog), GTK_RESPONSE_CLOSE, TRUE);
 		return;
 	}
 
@@ -471,12 +544,6 @@ static void send_notify(DBusGProxy *proxy,
 	first_update = get_system_time();
 }
 
-static void value_free(GValue *value)
-{
-	g_value_unset(value);
-	g_free(value);
-}
-
 static void
 select_device_changed(BluetoothChooser *sel,
 		      char *address,
@@ -582,9 +649,6 @@ static GOptionEntry options[] = {
 int main(int argc, char *argv[])
 {
 	DBusGProxy *proxy;
-	GHashTable *hash = NULL;
-	GValue *value;
-	ObexAgent *agent;
 	GError *error = NULL;
 	int i;
 
@@ -700,20 +764,14 @@ int main(int argc, char *argv[])
 	obex_agent_set_request_func(agent, request_callback, NULL);
 	obex_agent_set_progress_func(agent, progress_callback, NULL);
 	obex_agent_set_complete_func(agent, complete_callback, NULL);
+	obex_agent_set_error_func(agent, error_callback, NULL);
 
 	obex_agent_setup(agent, AGENT_PATH);
 
-	hash = g_hash_table_new_full(g_str_hash, g_str_equal,
-				     g_free, (GDestroyNotify) value_free);
-
-	value = g_new0(GValue, 1);
-	g_value_init(value, G_TYPE_STRING);
-	g_value_set_string(value, option_device);
-	g_hash_table_insert(hash, g_strdup("Destination"), value);
-
 	dbus_g_proxy_begin_call(proxy, "SendFiles",
 				send_notify, NULL, NULL,
-				dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), hash,
+				dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE),
+				new_hash_table_for_address (option_device),
 				G_TYPE_STRV, option_files,
 				DBUS_TYPE_G_OBJECT_PATH, AGENT_PATH,
 				G_TYPE_INVALID);
